import codecs
import functools
import importlib
import itertools
import logging
import os
import re
import sys
import uuid

import arrow
import cassandra
import cassandra.auth
import cassandra.cluster
import cassandra.policies
import tabulate

from . import cql
from . import exceptions
from . import migration as cstar_migration


CREATE_MIGRATIONS_TABLE = """
CREATE TABLE {keyspace}.{table} (
    id uuid,
    version int,
    name text,
    content text,
    checksum blob,
    state text,
    applied_at timestamp,
    PRIMARY KEY (id)
) WITH caching = {{'keys': 'NONE', 'rows_per_partition': 'NONE'}};
"""

CREATE_KEYSPACE = """
CREATE KEYSPACE {keyspace}
WITH REPLICATION = {replication}
AND DURABLE_WRITES = {durable_writes};
"""

DROP_KEYSPACE = """
DROP KEYSPACE IF EXISTS "{keyspace}";
"""

CREATE_DB_VERSION = """
INSERT INTO "{keyspace}"."{table}"
(id, version, name, content, checksum, state, applied_at)
VALUES (%s, %s, %s, %s, %s, %s, toTimestamp(now())) IF NOT EXISTS
"""

FINALIZE_DB_VERSION = """
UPDATE "{keyspace}"."{table}" SET state = %s WHERE id = %s IF state = %s
"""

DELETE_DB_VERSION = """
DELETE FROM "{keyspace}"."{table}" WHERE id = %s IF state = %s
"""


def cassandra_ddl_repr(data):
    """Generate a string representation of a map suitable for use in Cassandra DDL."""
    if isinstance(data, str):
        return "'" + re.sub(r"(?<!\\)'", "\\'", data) + "'"
    elif isinstance(data, dict):
        pairs = []
        for k, v in data.items():
            if not isinstance(k, str):
                raise ValueError('DDL map keys must be strings')

            pairs.append(cassandra_ddl_repr(k) + ': ' + cassandra_ddl_repr(v))
        return '{' + ', '.join(pairs) + '}'
    elif isinstance(data, int):
        return str(data)
    elif isinstance(data, bool):
        if data:
            return 'true'
        else:
            return 'false'
    else:
        raise ValueError('Cannot convert data to a DDL representation')


def confirmation_required(func):
    """Asks for the user's confirmation before calling the decorated function.

    This step is ignored when the script is not run from a TTY.
    """
    @functools.wraps(func)
    def wrapper(self, opts, *args, **kwargs):
        cli_mode = getattr(opts, 'cli_mode', False)
        if cli_mode and not opts.assume_yes:
            confirmation = input(f'The {func.__name__!r} operation cannot be undone. '
                                 'Are you sure? [y/N] ')
            if not confirmation.lower().startswith("y"):
                return
        opts.assume_yes = True
        return func(self, opts, *args, **kwargs)
    return wrapper


class Migrator(object):
    """Execute migration operations in a Cassandra database based on configuration.

    `opts` must contain at least the following attributes:
    - config_file: path to a YAML file containing the configuration
    - profiles: map of profile names to keyspace settings
    - user, password: authentication options. May be None to not use it.
    - hosts: comma-separated list of contact points
    - port: connection port
    """

    logger = logging.getLogger('Migrator')

    def __init__(self, config, profile='dev', hosts=['127.0.0.1'], port=9042,
                 user=None, password=None, protocol_version=4,
                 host_cert_path=None, client_key_path=None,
                 client_cert_path=None):
        self.config = config

        try:
            self.current_profile = self.config.profiles[profile]
        except KeyError:
            raise ValueError(f'Invalid profile name {profile!r}')

        if user:
            auth_provider = cassandra.auth.PlainTextAuthProvider(user, password)
        else:
            auth_provider = None

        if host_cert_path:
            ssl_options = self._build_ssl_options(
                host_cert_path,
                client_key_path,
                client_cert_path)
        else:
            ssl_options = None

        execution_profile = cassandra.cluster.ExecutionProfile(
            consistency_level=cassandra.ConsistencyLevel.ALL,
            load_balancing_policy=cassandra.policies.RoundRobinPolicy(),
            serial_consistency_level=cassandra.ConsistencyLevel.SERIAL,
            request_timeout=120,
        )
        execution_profiles = {
            cassandra.cluster.EXEC_PROFILE_DEFAULT: execution_profile,
        }

        self.cluster = cassandra.cluster.Cluster(
            contact_points=hosts,
            port=port,
            auth_provider=auth_provider,
            protocol_version=protocol_version,
            max_schema_agreement_wait=300,
            control_connection_timeout=10,
            connect_timeout=30,
            ssl_options=ssl_options,
            execution_profiles=execution_profiles,
        )

        self._session = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._session is not None:
            self._session.shutdown()
            self._session = None

        if self.cluster is not None:
            self.cluster.shutdown()
            self.cluster = None

    def _build_ssl_options(self, host_cert_path, client_key_path,
                           client_cert_path):
        return {
            'ca_certs': host_cert_path,
            'certfile': client_cert_path,
            'keyfile': client_key_path
        }

    def _check_cluster(self):
        """Check if the cluster is still alive, raise otherwise"""
        if not self.cluster:
            raise RuntimeError("Cluster has shut down")

    def _init_session(self):
        if not self._session:
            self._session = self.cluster.connect()

    @property
    def session(self):
        """Initialize and configure a  Cassandra driver session if needed."""

        self._check_cluster()
        self._init_session()

        return self._session

    def _get_target_version(self, v):
        """Parses a version specifier to an actual numeric migration version.

        `v` might be:
        - None: the latest version is chosen
        - an int, or a numeric string: that exact version is chosen
        - a string: the version with that name is chosen if it exists

        If an invalid version is found, `ValueError` is raised.
        """

        if v is None:
            return len(self.config.migrations)

        if isinstance(v, int):
            num = v
        elif v.isdigit():
            num = int(v)
        else:
            try:
                num = self.config.migrations.index(v)
            except IndexError:
                num = -1

        if num <= 0:
            raise ValueError('Invalid database version, must be a number > 0 '
                             'or the name of an existing migration')
        return num

    def _q(self, query, **kwargs):
        """Format a query with the configured keyspace and migration table.

        `keyspace` and `table` are interpolated as named arguments.
        """
        return query.format(
            keyspace=self.config.keyspace, table=self.config.migrations_table,
            **kwargs)

    def _execute(self, query, *args, **kwargs):
        """Execute a query with the current session."""

        self.logger.debug(f'Executing query: {query}')
        return self.session.execute(query, *args, **kwargs)

    def _keyspace_exists(self):
        self._init_session()

        return self.config.keyspace in self.cluster.metadata.keyspaces

    def _ensure_keyspace(self):
        """Create the keyspace if it does not exist."""

        if self._keyspace_exists():
            return

        self.logger.info(f'Creating keyspace {self.config.keyspace!r}')

        profile = self.current_profile
        self._execute(self._q(
            CREATE_KEYSPACE,
            replication=cassandra_ddl_repr(profile['replication']),
            durable_writes=cassandra_ddl_repr(profile['durable_writes'])))

        self.cluster.refresh_keyspace_metadata(self.config.keyspace)

    def _table_exists(self):
        self._init_session()

        ks_metadata = self.cluster.metadata.keyspaces.get(self.config.keyspace,
                                                          None)
        # Fail if the keyspace is missing. If it should be created
        # automatically _ensure_keyspace() must be called first.
        if not ks_metadata:
            raise ValueError(f'Keyspace {self.config.keyspace!r} does not exist, '
                             'stopping')

        return self.config.migrations_table in ks_metadata.tables

    def _ensure_table(self):
        """Create the migration table if it does not exist."""

        if self._table_exists():
            return

        self.logger.info(
            f'Creating table {self.config.migrations_table!r} '
            f'in keyspace {self.config.keyspace!r}'
        )

        self._execute(self._q(CREATE_MIGRATIONS_TABLE))
        self.cluster.refresh_table_metadata(self.config.keyspace,
                                            self.config.migrations_table)

    def _verify_migrations(self, migrations, ignore_failed=False,
                           ignore_concurrent=False):
        """Verify if the version history persisted in Cassandra matches the migrations.

        Migrations with corresponding DB versions must have the same content and name.
        Every DB version must have a corresponding migration. Migrations without
        corresponding DB versions are considered pending, and returned as a result.

        Returns a list of tuples of (version, migration), with version starting
        from 1 and incrementing linearly.
        """

        # Load all the currently existing versions and sort them by version
        # number, as Cassandra can only sort it for us by partition
        cur_versions = self._execute(
            self._q('SELECT * FROM "{keyspace}"."{table}"'),
        )
        cur_versions = sorted(cur_versions, key=lambda v: v.version)

        last_version = None
        version_pairs = itertools.zip_longest(cur_versions, migrations)

        # Work through ordered pairs of (existing version, expected_migration), so that
        # stored versions and expected migrations can be compared for any differences
        for i, (version, migration) in enumerate(version_pairs, 1):
            # If version is empty, the migration has not yet been applied.
            # Keep track of the first such version, and append it to the
            # pending migrations list.
            if not version:
                break

            # If migration is empty, we have a version in the database with
            # no corresponding file. That might mean we're running the wrong
            # migration or have an out-of-date state, so we must fail
            if not migration:
                raise exceptions.UnknownMigration(version.version, version.name)

            # A migration was previously run and failed
            if version.state == cstar_migration.MigrationState.FAILED:
                if ignore_failed:
                    break

                raise exceptions.FailedMigration(version.version, version.name)

            last_version = version.version

            # A migration is in progress
            if version.state == cstar_migration.MigrationState.IN_PROGRESS:
                if ignore_concurrent:
                    break
                raise exceptions.ConcurrentMigration(version.version, version.name)

            # A stored version's migrations differs from the one in the FS
            if (
                    version.content != migration.content or
                    version.name != migration.name or
                    bytearray(version.checksum) != bytearray(migration.checksum)
            ):
                raise exceptions.InconsistentState(migration, version)

        if not last_version:
            pending_migrations = list(migrations)
        else:
            pending_migrations = list(migrations)[last_version:]

        if not pending_migrations:
            self.logger.info('Database is already up-to-date')
        else:
            self.logger.info(
                f'Pending migrations found. Current version: {last_version}, '
                f'Latest version: {len(migrations)}'
            )

        pending_migrations = enumerate(
            pending_migrations, (last_version or 0) + 1)
        return last_version, cur_versions, list(pending_migrations)

    def _create_version(self, version, migration):
        """Write an in-progress version entry to Cassandra.

        The migration is inserted with the given `version` number if and only
        if it does not exist already (using a CompareAndSet operation).

        If the insert suceeds (with the migration marked as in-progress), we
        can continue and actually execute it. Otherwise, there was a concurrent
        write and we must fail to allow the other write to continue.

        """

        self.logger.info(f'Writing in-progress migration version {version}: {migration}')

        version_id = uuid.uuid4()
        result = self._execute(
            self._q(CREATE_DB_VERSION),
            (
                version_id,
                version,
                migration.name,
                migration.content,
                bytearray(migration.checksum),
                cstar_migration.MigrationState.IN_PROGRESS,
            ),
        )

        if not result or not result.one().applied:
            raise exceptions.ConcurrentMigration(version, migration.name)

        return version_id

    def _apply_cql_migration(self, version, migration):
        """Persist and apply a CQL migration.

        First create an in-progress version entry, apply the script, then
        finalize the entry as succeeded, failed or skipped.
        """

        self.logger.info('Applying CQL migration')

        statements = cql.CQLSplitter.split(migration.content)

        try:
            if statements:
                self.logger.info(
                    f'Executing migration with {len(statements)} CQL statements'
                )

            for statement in statements:
                self.session.execute(statement)
        except Exception:
            self.logger.exception('Failed to execute migration')
            raise exceptions.FailedMigration(version, migration.name)

    def _apply_python_migration(self, version, migration):
        """Persist and apply a python migration.

        First create an in-progress version entry, apply the script, then
        finalize the entry as succeeded, failed or skipped.
        """
        self.logger.info('Applying python script')

        try:
            mod, _ = os.path.splitext(os.path.basename(migration.path))
            migration_script = importlib.import_module(mod)
            migration_script.execute(self._session)
        except Exception:
            self.logger.exception('Failed to execute script')
            raise exceptions.FailedMigration(version, migration.name)

    def _apply_migration(self, version, migration, skip=False):
        """Persist and apply a migration.

        When `skip` is True, do everything but actually run the script, for
        example, when baselining instead of migrating.
        """

        self.logger.info(f'Advancing to version {version}')

        version_uuid = self._create_version(version, migration)
        new_state = cstar_migration.MigrationState.FAILED
        sys.path.append(self.config.migrations_path)

        result = None

        try:
            if skip:
                self.logger.info('Migration is marked for skipping, '
                                 'not actually running script')
            else:
                if migration.is_python:
                    self._apply_python_migration(version, migration)
                else:
                    self._apply_cql_migration(version, migration)
        except Exception:
            self.logger.exception('Failed to execute migration')
            raise exceptions.FailedMigration(version, migration.name)
        else:
            new_state = (cstar_migration.MigrationState.SUCCEEDED if not skip
                         else cstar_migration.MigrationState.SKIPPED)
        finally:
            self.logger.info(f'Finalizing migration version with state {new_state!r}')
            result = self._execute(
                self._q(FINALIZE_DB_VERSION),
                (new_state, version_uuid, cstar_migration.MigrationState.IN_PROGRESS))

        if not result or not result.one().applied:
            raise exceptions.ConcurrentMigration(version, migration.name)

    def _cleanup_previous_versions(self, cur_versions):
        if not cur_versions:
            return

        last_version = cur_versions[-1]
        if last_version.state != cstar_migration.MigrationState.FAILED:
            return

        self.logger.warn(
            f'Cleaning up previous failed migration '
            f'(version {last_version.version}): {last_version.name}'
        )

        result = self._execute(
            self._q(DELETE_DB_VERSION),
            (last_version.id, cstar_migration.MigrationState.FAILED),
        )

        if not result.one().applied:
            raise exceptions.ConcurrentMigration(last_version.version,
                                                 last_version.name)

    def _advance(self, migrations, target, cur_versions, skip=False, force=False):
        """Apply all necessary migrations to reach a target version."""
        if force:
            self._cleanup_previous_versions(cur_versions)

        target_version = self._get_target_version(target)

        if migrations:
            # Set default keyspace so migrations don't need to refer to it manually
            self.session.execute(f'USE {self.config.keyspace};')

        for version, migration in migrations:
            if version > target_version:
                break

            self._apply_migration(version, migration, skip=skip)

        self.cluster.refresh_schema_metadata()

    @confirmation_required
    def migrate(self, opts):
        """Migrate a database to a given version, applying any needed migration."""

        self._check_cluster()

        self._ensure_keyspace()
        self._ensure_table()

        last_version, cur_versions, pending_migrations = \
            self._verify_migrations(self.config.migrations,
                                    ignore_failed=opts.force)

        self._advance(pending_migrations, opts.db_version, cur_versions,
                      force=opts.force)

    @confirmation_required
    def reset(self, opts):
        """Reset the database, by dropping an existing keyspace then running a migration.
        """
        self._check_cluster()

        self.logger.info(f'Dropping existing keyspace {self.config.keyspace!r}')

        self._execute(self._q(DROP_KEYSPACE))
        self.cluster.refresh_schema_metadata()

        opts.force = False
        self.migrate(opts)

    @confirmation_required
    def clear(self, opts):
        """Clear the database, by dropping an existing keyspace."""
        self._check_cluster()

        self.logger.info(f'Dropping existing keyspace {self.config.keyspace!r}')

        self._execute(self._q(DROP_KEYSPACE))
        self.cluster.refresh_schema_metadata()

    @confirmation_required
    def baseline(self, opts):
        """Baseline a database, by advancing migration state without changes."""

        self._check_cluster()
        self._ensure_table()

        last_version, cur_versions, pending_migrations = \
            self._verify_migrations(self.config.migrations,
                                    ignore_failed=False)

        self._advance(pending_migrations, opts.db_version, cur_versions,
                      skip=True)

    @staticmethod
    def _bytes_to_hex(bs):
        return codecs.getencoder('hex')(bs)[0]

    def status(self, opts):
        self._check_cluster()

        if not self._keyspace_exists():
            print(f'Keyspace {self.config.keyspace!r} does not exist')
            return

        if not self._table_exists():
            print(
                f'Migration table {self.config.migrations_table!r} does not exist in '
                f'keyspace {self.config.keyspace!r}'
            )
            return

        last_version, cur_versions, pending_migrations = self._verify_migrations(
            self.config.migrations, ignore_failed=True, ignore_concurrent=True)
        latest_version = len(self.config.migrations)

        print(tabulate.tabulate(
            (
                ('Keyspace:', self.config.keyspace),
                ('Migrations table:', self.config.migrations_table),
                ('Current DB version:', last_version),
                ('Latest DB version:', latest_version),
            ),
            tablefmt='plain'))

        if cur_versions:
            print('\n## Applied migrations\n')

            data = []
            for version in cur_versions:
                checksum = self._bytes_to_hex(version.checksum)
                date = arrow.get(version.applied_at).format()
                data.append(
                    (str(version.version), version.name, version.state, date, checksum),
                )
            print(tabulate.tabulate(data, headers=['#', 'Name', 'State',
                                                   'Date applied', 'Checksum']))

        if pending_migrations:
            print('\n## Pending migrations\n')

            data = []
            for version, migration in pending_migrations:
                checksum = self._bytes_to_hex(migration.checksum)
                data.append(
                    (str(version), migration.name, checksum),
                )
            print(tabulate.tabulate(data, headers=['#', 'Name', 'Checksum']))



import collections
import glob
import hashlib
import io
import os
import re

import arrow


class MigrationState:
    """Possible states of a migration, as saved in Cassandra."""

    SUCCEEDED = 'SUCCEEDED'
    FAILED = 'FAILED'
    SKIPPED = 'SKIPPED'
    IN_PROGRESS = 'IN_PROGRESS'


class Migration(collections.namedtuple('Migration',
                                       'path name is_python content checksum')):
    """Data class representing the specification of a migration.

    Migrations can take the form of CQL files or Python scripts, and usually
    have names starting with a version string that can be ordered.
    A checksum is kept to allow detecting changes to previously applied
    migrations"""

    __slots__ = ()

    @staticmethod
    def _natural_sort_key(s):
        """Generate a sort key for natural sorting."""
        k = tuple(int(text) if text.isdigit() else text
                  for text in re.split(r'([0-9]+)', s))
        return k

    @classmethod
    def load(cls, path):
        """Load a migration from a given file."""
        with open(path, 'r', encoding='utf-8') as fp:
            content = fp.read()

        checksum = bytes(hashlib.sha256(content.encode('utf-8')).digest())
        # Should use enum but python3 requires importing an extra library
        # Reconsidering the use of enums. This is a binary decision.
        # Boolean will work just fine.
        is_python = bool(re.findall(r"\.(py)$", os.path.abspath(path)))

        return cls(os.path.abspath(path), os.path.basename(path),
                   is_python, content, checksum)

    @classmethod
    def sort_paths(cls, paths):
        """Sort paths naturally by basename, to order by migration version."""
        return sorted(paths,
                      key=lambda p: cls._natural_sort_key(os.path.basename(p)))

    @classmethod
    def glob_all(cls, base_path, *patterns):
        """Load all paths matching a glob as migrations in sorted order."""

        paths = []
        for pattern in patterns:
            paths.extend(glob.iglob(os.path.join(base_path, pattern)))

        return list(map(cls.load, cls.sort_paths(paths)))

    @classmethod
    def generate(cls, config, description, output):
        fname_fmt = config.new_migration_name
        text_cql_fmt = config.new_cql_migration_text
        text_py_fmt = config.new_python_migration_text

        clean_desc = re.sub(r'[\W\s]+', '_', description)
        next_version = len(config.migrations) + 1
        date = arrow.utcnow()

        format_args = {
            'desc': clean_desc.lower(),
            'full_desc': description,
            'next_version': next_version,
            'date': date,
            'keyspace': config.keyspace
        }

        if output == "python":
            file_extension = ".py"
            file_content = text_py_fmt.format(**format_args)
        else:
            file_extension = ".cql"
            file_content = text_cql_fmt.format(**format_args)

        fname = fname_fmt.format(**format_args) + file_extension
        new_path = os.path.join(config.migrations_path, fname)

        cls._create_file(new_path, file_content)

        return new_path

    @classmethod
    def _create_file(cls, path, content):
        """Creates physical file."""
        with io.open(path, 'w', encoding='utf-8') as f:
            f.write(content + '\n')

    def __str__(self):
        return f'Migration({self.name!r})'



import pytest

import cstarmigrate.cql


@pytest.mark.parametrize('cql,statements', [
    # Two statements, with whitespace
    (
        """
        CREATE TABLE hello;
        CREATE TABLE world;
        """,
        ['CREATE TABLE hello', 'CREATE TABLE world'],
    ),

    # Two statements, no whitespace
    (
        """
        CREATE TABLE hello;
        CREATE TABLE world;
        """,
        ['CREATE TABLE hello', 'CREATE TABLE world'],
    ),

    # Two statements, with line and block comments
    (
        """
        // comment
        -- comment
        CREATE TABLE hello;
        /* comment; comment
        */
        CREATE TABLE world;
        """,
        ['CREATE TABLE hello', 'CREATE TABLE world']
    ),

    # Statements with semicolons inside strings
    (
        """
        CREATE TABLE 'hello;';
        CREATE TABLE "world;"
        """,
        ["CREATE TABLE 'hello;'", 'CREATE TABLE "world;"'],
    ),

    # Double-dollar-sign quoted strings
    (
        """
        INSERT INTO test (test)
        VALUES ($$Pesky semicolon here ;Hello$$);
        """,
        ["INSERT INTO test (test) VALUES ($$Pesky semicolon here ;Hello$$)"],
    ),
])
def test_cql_split(cql, statements):
    result = cstarmigrate.cql.CQLSplitter.split(cql.strip())

    assert result == statements



import argparse
import logging
import os
import subprocess
import sys

from . import config as cstar_config
from . import exceptions
from . import migration as cstar_migration
from . import migrator as cstar_migrator


def open_file(filename):
    if sys.platform == 'win32':
        os.startfile(filename)
    else:
        if 'XDG_CURRENT_DESKTOP' in os.environ:
            opener = ['xdg-open']
        elif 'EDITOR' in os.environ:
            opener = [os.environ['EDITOR']]
        else:
            opener = ['vi']

        opener.append(filename)
        subprocess.call(opener)


def main(argv=None):
    logging.basicConfig(level=logging.INFO)
    logging.getLogger('cassandra.policies').setLevel(logging.ERROR)

    parser = argparse.ArgumentParser(description='Cassandra schema migration tool')
    parser.add_argument('-H', '--hosts', default='127.0.0.1',
                        help='Comma-separated list of contact points')
    parser.add_argument('-p', '--port', type=int, default=9042,
                        help='Connection port')
    parser.add_argument('-u', '--user',
                        help='Connection username')
    parser.add_argument('-P', '--password',
                        help='Connection password')
    parser.add_argument('-l', '--protocol-version', type=int, default=4,
                        help='Connection protocol version')
    parser.add_argument('-c', '--config-file', default='cstar-migrate.yml',
                        help='Path to configuration file')
    parser.add_argument('-m', '--profile', default='dev',
                        help='Name of keyspace profile to use')
    parser.add_argument('-s', '--ssl-cert', default=None,
                        help="""
                        File path of .pem or .crt containing certificate of the
                        cassandra host you are connecting to (or the
                        certificate of the CA that signed the host certificate).
                        If this option is provided, cstar-migrate will use
                        ssl to connect to the cluster. If this option is not
                        provided, the -k and -t options will be ignored. """)
    parser.add_argument('-k', '--ssl-client-private-key', default=None,
                        help="""
                        File path of the .key file containing the private key
                        of the host on which the cstar-migrate command is
                        run. This option must be used in conjuction with the
                        -t option. This option is ignored unless the -s
                        option is provided.""")
    parser.add_argument('-t', '--ssl-client-cert', default=None,
                        help="""
                        File path of the .crt file containing the public
                        certificate of the host on which the cstar-migrate
                        command is run. This certificate (or the CA that signed
                        it) must be trusted by the cassandra host that
                        migrations are run against. This option must be used in
                        conjuction with the -k option. This option is ignored
                        unless the -s option is provided.""")
    parser.add_argument('-y', '--assume-yes', action='store_true',
                        help='Automatically answer "yes" for all questions')

    commands = parser.add_subparsers(help='sub-command help')

    migrate = commands.add_parser(
        'migrate',
        help='Migrate database up to the most recent (or specified) version '
             'by applying any new migration scripts in sequence')
    migrate.add_argument('-f', '--force', action='store_true',
                         help='Force migration even if last attempt failed')
    migrate.set_defaults(action='migrate')

    reset = commands.add_parser(
        'reset',
        help='Reset the database, by dropping an existing keyspace, '
             'then running a migration')
    reset.set_defaults(action='reset')

    clear = commands.add_parser(
        'clear',
        help='Clear the database, by dropping an existing keyspace')
    clear.set_defaults(action='clear')

    baseline = commands.add_parser(
        'baseline',
        help='Baseline database state, advancing migration information without '
             'making changes')
    baseline.set_defaults(action='baseline')

    stats = commands.add_parser(
        'status',
        help='Print current state of keyspace')
    stats.set_defaults(action='status')

    generate = commands.add_parser(
        'generate',
        help='Generate a new migration file')
    generate.add_argument(
        'description',
        help='Brief description of the new migration')
    generate.add_argument(
        '--python',
        dest='migration_type',
        action='store_const',
        const='python',
        default='cql',
        help='Generates a Python script instead of CQL.')

    generate.set_defaults(action='generate')

    for subcommand in (migrate, reset, baseline):
        subcommand.add_argument('db_version', metavar='VERSION', nargs='?',
                                help='Database version to baseline/reset/migrate to')

    opts = parser.parse_args(argv)

    # Enable user confirmation if we're running the script from a TTY
    opts.cli_mode = sys.stdin.isatty()

    config = cstar_config.MigrationConfig.load(opts.config_file)

    # Print help by default if no action was provided
    if getattr(opts, 'action', None) is None:
        parser.print_help()

        sys.exit(0)

    # Provide ability to open generated migration when executed from a TTY
    if opts.action == 'generate':
        new_path = cstar_migration.Migration.generate(
            config=config,
            description=opts.description,
            output=opts.migration_type,
        )
        if sys.stdin.isatty():
            open_file(new_path)

        print(os.path.basename(new_path))

        sys.exit(0)

    # Handle the rest
    with cstar_migrator.Migrator(
        config=config,
        profile=opts.profile,
        hosts=opts.hosts.split(','),
        port=opts.port,
        user=opts.user,
        password=opts.password,
        protocol_version=opts.protocol_version,
        host_cert_path=opts.ssl_cert,
        client_key_path=opts.ssl_client_private_key,
        client_cert_path=opts.ssl_client_cert,
    ) as migrator:
        cmd_method = getattr(migrator, opts.action)
        if not callable(cmd_method):
            print('Error: invalid command', file=sys.stderr)
            sys.exit(1)

        try:
            cmd_method(opts)
        except exceptions.MigrationError as e:
            print(f'Error: {e!r}', file=sys.stderr)
            sys.exit(1)



class MigrationError(RuntimeError):
    """Base class for migration errors."""
    pass


class FailedMigration(MigrationError):
    """Database state contains failed migrations."""

    def __init__(self, version, name):
        self.version = version
        self.migration_name = name

        super(FailedMigration, self).__init__(
            f'Migration failed, cannot continue (version {version}): {name}')


class ConcurrentMigration(MigrationError):
    """Database state contains failed migrations."""

    def __init__(self, version, name):
        self.version = version
        self.migration_name = name

        super(ConcurrentMigration, self).__init__(
            f'Migration already in progress (version {version}): {name}')


class InconsistentState(MigrationError):
    """Database state differs from specified migrations."""

    def __init__(self, migration, version):
        self.migration = migration
        self.version = version

        super(InconsistentState, self).__init__(
            f'Found inconsistency between specified migration and stored '
            f'version: {migration} != {version}')


class UnknownMigration(MigrationError):
    """Database contains migrations that have not been specified."""
    def __init__(self, version, name):
        self.version = version
        self.migration_name = name

        super(UnknownMigration, self).__init__(
            f'Found version in database without corresponding '
            f'migration (version {version}): {name}')
